package solutions.bitbadger.documents.core.tests.integration

import solutions.bitbadger.documents.Configuration
import solutions.bitbadger.documents.Dialect
import solutions.bitbadger.documents.Field
import solutions.bitbadger.documents.FieldMatch
import solutions.bitbadger.documents.core.tests.ArrayDocument
import solutions.bitbadger.documents.core.tests.JsonDocument
import solutions.bitbadger.documents.core.tests.NumIdDocument
import solutions.bitbadger.documents.core.tests.TEST_TABLE
import solutions.bitbadger.documents.java.extensions.*
import kotlin.test.assertEquals
import kotlin.test.assertTrue

/**
 * Tests for the JSON-returning functions
 *
 * NOTE: PostgreSQL JSONB columns do not preserve the original JSON with which a document was stored. These tests are
 * the most complex within the library, as they have split testing based on the backing data store. The PostgreSQL tests
 * check IDs (and, in the case of ordered queries, which ones occur before which others) vs. the entire JSON string.
 * Meanwhile, SQLite stores JSON as text, and will return exactly the JSON it was given when it was originally written.
 * These tests can ensure the expected round-trip of the entire JSON string.
 */
object JsonFunctions {

    /**
     * PostgreSQL, when returning JSONB as a string, has spaces after commas and colons delineating fields and values.
     * This function will do a crude string replacement to match the target string based on the dialect being tested.
     *
     * @param json The JSON which should be returned
     * @return The actual expected JSON based on the database being tested
     */
    fun maybeJsonB(json: String) =
        when (Configuration.dialect()) {
            Dialect.SQLITE -> json
            Dialect.POSTGRESQL -> json.replace("\":", "\": ").replace(",\"", ", \"")
        }

    /**
     * Create a snippet of JSON to find a document ID
     *
     * @param id The ID of the document
     * @return A connection-aware ID to check for presence and positioning
     */
    private fun docId(id: String) =
        maybeJsonB("{\"id\":\"$id\"")

    fun allDefault(db: ThrowawayDatabase) {
        JsonDocument.load(db)
        val json = db.conn.jsonAll(TEST_TABLE)
        assertTrue(json.startsWith("["), "JSON should start with '[' ($json)")
        when (Configuration.dialect()) {
            Dialect.SQLITE -> {
                assertTrue(json.contains(JsonDocument.one), "Document 'one' not found in JSON ($json)")
                assertTrue(json.contains(JsonDocument.two), "Document 'two' not found in JSON ($json)")
                assertTrue(json.contains(JsonDocument.three), "Document 'three' not found in JSON ($json)")
                assertTrue(json.contains(JsonDocument.four), "Document 'four' not found in JSON ($json)")
                assertTrue(json.contains(JsonDocument.five), "Document 'five' not found in JSON ($json)")
            }
            Dialect.POSTGRESQL -> {
                assertTrue(json.indexOf(docId("one")) >= 0, "Document 'one' not found in JSON ($json)")
                assertTrue(json.indexOf(docId("two")) >= 0, "Document 'two' not found in JSON ($json)")
                assertTrue(json.indexOf(docId("three")) >= 0, "Document 'three' not found in JSON ($json)")
                assertTrue(json.indexOf(docId("four")) >= 0, "Document 'four' not found in JSON ($json)")
                assertTrue(json.indexOf(docId("five")) >= 0, "Document 'five' not found in JSON ($json)")
            }
        }
        assertTrue(json.endsWith("]"), "JSON should end with ']' ($json)")
    }

    fun allEmpty(db: ThrowawayDatabase) =
        assertEquals("[]", db.conn.jsonAll(TEST_TABLE), "There should have been no documents returned")

    fun byIdString(db: ThrowawayDatabase) {
        JsonDocument.load(db)
        val json = db.conn.jsonById(TEST_TABLE, "two")
        when (Configuration.dialect()) {
            Dialect.SQLITE -> assertEquals(JsonDocument.two, json, "An incorrect document was returned")
            Dialect.POSTGRESQL -> assertTrue(
                json.indexOf(docId("two")) >= 0,
                "An incorrect document was returned ($json)"
            )
        }
    }

    fun byIdNumber(db: ThrowawayDatabase) {
        Configuration.idField = "key"
        try {
            db.conn.insert(TEST_TABLE, NumIdDocument(18, "howdy"))
            assertEquals(
                maybeJsonB("{\"key\":18,\"text\":\"howdy\"}"), db.conn.jsonById(TEST_TABLE, 18),
                "The document should have been found by numeric ID"
            )
        } finally {
            Configuration.idField = "id"
        }
    }

    fun byIdNotFound(db: ThrowawayDatabase) {
        JsonDocument.load(db)
        assertEquals("{}", db.conn.jsonById(TEST_TABLE, "x"), "There should have been no document returned")
    }

    fun byFieldsMatch(db: ThrowawayDatabase) {
        JsonDocument.load(db)
        val json = db.conn.jsonByFields(
            TEST_TABLE, listOf(Field.any("value", listOf("blue", "purple")), Field.exists("sub")), FieldMatch.ALL
        )
        when (Configuration.dialect()) {
            Dialect.SQLITE -> assertEquals("[${JsonDocument.four}]", json, "The incorrect document was returned")
            Dialect.POSTGRESQL -> {
                assertTrue(json.startsWith("["), "JSON should start with '[' ($json)")
                assertTrue(json.contains(docId("four")),"The incorrect document was returned ($json)")
                assertTrue(json.endsWith("]"), "JSON should end with ']' ($json)")
            }
        }
    }

    fun byFieldsMatchOrdered(db: ThrowawayDatabase) {
        JsonDocument.load(db)
        val json = db.conn.jsonByFields(
            TEST_TABLE, listOf(Field.equal("value", "purple")), orderBy = listOf(Field.named("id"))
        )
        when (Configuration.dialect()) {
            Dialect.SQLITE -> assertEquals(
                "[${JsonDocument.five},${JsonDocument.four}]", json, "The documents were not ordered correctly"
            )

            Dialect.POSTGRESQL -> {
                val fiveIdx = json.indexOf(docId("five"))
                val fourIdx = json.indexOf(docId("four"))
                assertTrue(json.startsWith("["), "JSON should start with '[' ($json)")
                assertTrue(fiveIdx >= 0, "Document 'five' not found ($json)")
                assertTrue(fourIdx >= 0, "Document 'four' not found ($json)")
                assertTrue(fiveIdx < fourIdx, "Document 'five' should have been before 'four' ($json)")
                assertTrue(json.endsWith("]"), "JSON should end with ']' ($json)")
            }
        }
    }

    fun byFieldsMatchNumIn(db: ThrowawayDatabase) {
        JsonDocument.load(db)
        val json = db.conn.jsonByFields(TEST_TABLE, listOf(Field.any("numValue", listOf(2, 4, 6, 8))))
        when (Configuration.dialect()) {
            Dialect.SQLITE -> assertEquals("[${JsonDocument.three}]", json, "The incorrect document was returned")
            Dialect.POSTGRESQL -> {
                assertTrue(json.startsWith("["), "JSON should start with '[' ($json)")
                assertTrue(json.contains(docId("three")), "The incorrect document was returned ($json)")
                assertTrue(json.endsWith("]"), "JSON should end with ']' ($json)")
            }
        }
    }

    fun byFieldsNoMatch(db: ThrowawayDatabase) {
        JsonDocument.load(db)
        assertEquals(
            "[]", db.conn.jsonByFields(TEST_TABLE, listOf(Field.greater("numValue", 100))),
            "There should have been no documents returned"
        )
    }

    fun byFieldsMatchInArray(db: ThrowawayDatabase) {
        ArrayDocument.testDocuments.forEach { db.conn.insert(TEST_TABLE, it) }
        val json = db.conn.jsonByFields(TEST_TABLE, listOf(Field.inArray("values", TEST_TABLE, listOf("c"))))
        assertTrue(json.startsWith("["), "JSON should start with '[' ($json)")
        assertTrue(json.contains(docId("first")), "The 'first' document was not found ($json)")
        assertTrue(json.contains(docId("second")), "The 'second' document was not found ($json)")
        assertTrue(json.endsWith("]"), "JSON should end with ']' ($json)")
    }

    fun byFieldsNoMatchInArray(db: ThrowawayDatabase) {
        ArrayDocument.testDocuments.forEach { db.conn.insert(TEST_TABLE, it) }
        assertEquals(
            "[]", db.conn.jsonByFields(
                TEST_TABLE, listOf(Field.inArray("values", TEST_TABLE, listOf("j")))
            ), "There should have been no documents returned"
        )
    }

    fun byContainsMatch(db: ThrowawayDatabase) {
        JsonDocument.load(db)
        val json = db.conn.jsonByContains(TEST_TABLE, mapOf("value" to "purple"))
        assertTrue(json.startsWith("["), "JSON should start with '[' ($json)")
        when (Configuration.dialect()) {
            Dialect.SQLITE -> {
                assertTrue(json.contains(JsonDocument.four), "Document 'four' not found ($json)")
                assertTrue(json.contains(JsonDocument.five), "Document 'five' not found ($json)")
            }
            Dialect.POSTGRESQL -> {
                assertTrue(json.contains(docId("four")), "Document 'four' not found ($json)")
                assertTrue(json.contains(docId("five")), "Document 'five' not found ($json)")
            }
        }
        assertTrue(json.endsWith("]"), "JSON should end with ']' ($json)")
    }

    fun byContainsMatchOrdered(db: ThrowawayDatabase) {
        JsonDocument.load(db)
        val json = db.conn.jsonByContains(
            TEST_TABLE, mapOf("sub" to mapOf("foo" to "green")), listOf(Field.named("value"))
        )
        when (Configuration.dialect()) {
            Dialect.SQLITE -> assertEquals(
                "[${JsonDocument.two},${JsonDocument.four}]", json, "The documents were not ordered correctly"
            )
            Dialect.POSTGRESQL -> {
                val twoIdx = json.indexOf(docId("two"))
                val fourIdx = json.indexOf(docId("four"))
                assertTrue(json.startsWith("["), "JSON should start with '[' ($json)")
                assertTrue(twoIdx >= 0, "Document 'two' not found ($json)")
                assertTrue(fourIdx >= 0, "Document 'four' not found ($json)")
                assertTrue(twoIdx < fourIdx, "Document 'two' should have been before 'four' ($json)")
                assertTrue(json.endsWith("]"), "JSON should end with ']' ($json)")
            }
        }
    }

    fun byContainsNoMatch(db: ThrowawayDatabase) {
        JsonDocument.load(db)
        assertEquals(
            "[]",
            db.conn.jsonByContains(TEST_TABLE, mapOf("value" to "indigo")),
            "There should have been no documents returned"
        )
    }

    fun byJsonPathMatch(db: ThrowawayDatabase) {
        JsonDocument.load(db)
        val json = db.conn.jsonByJsonPath(TEST_TABLE, "$.numValue ? (@ > 10)")
        assertTrue(json.startsWith("["), "JSON should start with '[' ($json)")
        when (Configuration.dialect()) {
            Dialect.SQLITE -> {
                assertTrue(json.contains(JsonDocument.four), "Document 'four' not found ($json)")
                assertTrue(json.contains(JsonDocument.five), "Document 'five' not found ($json)")
            }
            Dialect.POSTGRESQL -> {
                assertTrue(json.contains(docId("four")), "Document 'four' not found ($json)")
                assertTrue(json.contains(docId("five")), "Document 'five' not found ($json)")
            }
        }
        assertTrue(json.endsWith("]"), "JSON should end with ']' ($json)")
    }

    fun byJsonPathMatchOrdered(db: ThrowawayDatabase) {
        JsonDocument.load(db)
        val json = db.conn.jsonByJsonPath(TEST_TABLE, "$.numValue ? (@ > 10)", listOf(Field.named("id")))
        when (Configuration.dialect()) {
            Dialect.SQLITE -> assertEquals(
                "[${JsonDocument.five},${JsonDocument.four}]", json, "The documents were not ordered correctly"
            )

            Dialect.POSTGRESQL -> {
                val fiveIdx = json.indexOf(docId("five"))
                val fourIdx = json.indexOf(docId("four"))
                assertTrue(json.startsWith("["), "JSON should start with '[' ($json)")
                assertTrue(fiveIdx >= 0, "Document 'five' not found ($json)")
                assertTrue(fourIdx >= 0, "Document 'four' not found ($json)")
                assertTrue(fiveIdx < fourIdx, "Document 'five' should have been before 'four' ($json)")
                assertTrue(json.endsWith("]"), "JSON should end with ']' ($json)")
            }
        }
    }

    fun byJsonPathNoMatch(db: ThrowawayDatabase) {
        JsonDocument.load(db)
        assertEquals(
            "[]",
            db.conn.jsonByJsonPath(TEST_TABLE, "$.numValue ? (@ > 100)"),
            "There should have been no documents returned"
        )
    }

    fun firstByFieldsMatchOne(db: ThrowawayDatabase) {
        JsonDocument.load(db)
        val json = db.conn.jsonFirstByFields(TEST_TABLE, listOf(Field.equal("value", "another")))
        when (Configuration.dialect()) {
            Dialect.SQLITE -> assertEquals(JsonDocument.two, json, "The incorrect document was returned")
            Dialect.POSTGRESQL -> assertTrue(json.contains(docId("two")), "The incorrect document was returned ($json)")
        }
    }

    fun firstByFieldsMatchMany(db: ThrowawayDatabase) {
        JsonDocument.load(db)
        val json = db.conn.jsonFirstByFields(TEST_TABLE, listOf(Field.equal("sub.foo", "green")))
        when (Configuration.dialect()) {
            Dialect.SQLITE -> assertTrue(
                json.contains(JsonDocument.two) || json.contains(JsonDocument.four),
                "Expected document 'two' or 'four' ($json)"
            )
            Dialect.POSTGRESQL -> assertTrue(
                json.contains(docId("two")) || json.contains(docId("four")),
                "Expected document 'two' or 'four' ($json)"
            )
        }
    }

    fun firstByFieldsMatchOrdered(db: ThrowawayDatabase) {
        JsonDocument.load(db)
        val json = db.conn.jsonFirstByFields(
            TEST_TABLE, listOf(Field.equal("sub.foo", "green")), orderBy = listOf(Field.named("n:numValue DESC"))
        )
        when (Configuration.dialect()) {
            Dialect.SQLITE -> assertEquals(JsonDocument.four, json, "An incorrect document was returned")
            Dialect.POSTGRESQL -> assertTrue(json.contains(docId("four")), "An incorrect document was returned ($json)")
        }
    }

    fun firstByFieldsNoMatch(db: ThrowawayDatabase) {
        JsonDocument.load(db)
        assertEquals(
            "{}",
            db.conn.jsonFirstByFields(TEST_TABLE, listOf(Field.equal("value", "absent"))),
            "There should have been no document returned"
        )
    }

    fun firstByContainsMatchOne(db: ThrowawayDatabase) {
        JsonDocument.load(db)
        val json = db.conn.jsonFirstByContains(TEST_TABLE, mapOf("value" to "FIRST!"))
        when (Configuration.dialect()) {
            Dialect.SQLITE -> assertEquals(JsonDocument.one, json, "An incorrect document was returned")
            Dialect.POSTGRESQL -> assertTrue(json.contains(docId("one")), "An incorrect document was returned ($json)")
        }
    }

    fun firstByContainsMatchMany(db: ThrowawayDatabase) {
        JsonDocument.load(db)
        val json = db.conn.jsonFirstByContains(TEST_TABLE, mapOf("value" to "purple"))
        when (Configuration.dialect()) {
            Dialect.SQLITE -> assertTrue(
                json.contains(JsonDocument.four) || json.contains(JsonDocument.five),
                "Expected document 'four' or 'five' ($json)"
            )
            Dialect.POSTGRESQL -> assertTrue(
                json.contains(docId("four")) || json.contains(docId("five")),
                "Expected document 'four' or 'five' ($json)"
            )
        }
    }

    fun firstByContainsMatchOrdered(db: ThrowawayDatabase) {
        JsonDocument.load(db)
        val json = db.conn.jsonFirstByContains(
            TEST_TABLE, mapOf("value" to "purple"), listOf(Field.named("sub.bar NULLS FIRST"))
        )
        when (Configuration.dialect()) {
            Dialect.SQLITE -> assertEquals(JsonDocument.five, json, "An incorrect document was returned")
            Dialect.POSTGRESQL -> assertTrue(json.contains(docId("five")), "An incorrect document was returned ($json)")
        }
    }

    fun firstByContainsNoMatch(db: ThrowawayDatabase) {
        JsonDocument.load(db)
        assertEquals(
            "{}",
            db.conn.jsonFirstByContains(TEST_TABLE, mapOf("value" to "indigo")),
            "There should have been no document returned"
        )
    }

    fun firstByJsonPathMatchOne(db: ThrowawayDatabase) {
        JsonDocument.load(db)
        val json = db.conn.jsonFirstByJsonPath(TEST_TABLE, "$.numValue ? (@ == 10)")
        when (Configuration.dialect()) {
            Dialect.SQLITE -> assertEquals(JsonDocument.two, json, "An incorrect document was returned")
            Dialect.POSTGRESQL -> assertTrue(json.contains(docId("two")), "An incorrect document was returned ($json)")
        }
    }

    fun firstByJsonPathMatchMany(db: ThrowawayDatabase) {
        JsonDocument.load(db)
        val json = db.conn.jsonFirstByJsonPath(TEST_TABLE, "$.numValue ? (@ > 10)")
        when (Configuration.dialect()) {
            Dialect.SQLITE -> assertTrue(
                json.contains(JsonDocument.four) || json.contains(JsonDocument.five),
                "Expected document 'four' or 'five' ($json)"
            )
            Dialect.POSTGRESQL -> assertTrue(
                json.contains(docId("four")) || json.contains(docId("five")),
                "Expected document 'four' or 'five' ($json)"
            )
        }
    }

    fun firstByJsonPathMatchOrdered(db: ThrowawayDatabase) {
        JsonDocument.load(db)
        val json = db.conn.jsonFirstByJsonPath(TEST_TABLE, "$.numValue ? (@ > 10)", listOf(Field.named("id DESC")))
        when (Configuration.dialect()) {
            Dialect.SQLITE -> assertEquals(JsonDocument.four, json, "An incorrect document was returned")
            Dialect.POSTGRESQL -> assertTrue(json.contains(docId("four")), "An incorrect document was returned ($json)")
        }
    }

    fun firstByJsonPathNoMatch(db: ThrowawayDatabase) {
        JsonDocument.load(db)
        assertEquals(
            "{}",
            db.conn.jsonFirstByJsonPath(TEST_TABLE, "$.numValue ? (@ > 100)"),
            "There should have been no document returned"
        )
    }
}
